from typing import List, TYPE_CHECKING

from .. import defaults
from . import subnet as base_subnet

if TYPE_CHECKING:
    from .network import Network
    from .subnet import BaseSubnet


class BaseNode:
    """Base class for Nodes representing computers in virtual networks"""
    lock_filepath = '/var/run/l3ns.lock'

    def __init__(self, name):
        """Create node

        Create basic node with black configuration.

        Args:
             name: unique identifier for new node
        """
        self.name = name
        self._interfaces = {}
        self._routes = {}
        self._networks: List['Network'] = []
        self._files = {}

        self.loaded = False
        self.started = False
        self.stopped = False
        self.is_router = False
        self._gateways = []
        self.connect_to_internet = False

    @property
    def is_gateway(self):
        return bool(self._gateways)

    def connect_to(self,
                   other_node: 'BaseNode',
                   subnet_name: str = None,
                   subnet_class: type = None,
                   **subnet_kwargs):
        """Create subnet that connects this node to another

        Create a subnet with that will consist of two nodes. You can specify
        specific subnet class and name. If not, name will be generated automatically
        and subnet class will be retrieved from [defaults](l3ns/defaults).

        Args:
            other_node: Node to connect
            subnet_name: Name to use for net subnet
            subnet_class: Class to use for new subnet
        """

        if subnet_class is None:
            subnet_class = defaults.subnet_class

        if subnet_name is None:
            subnet_name = self.name + '_' + other_node.name

        return subnet_class(subnet_name, self, other_node, **subnet_kwargs)

    def add_interface(self, ip_address, subnet: 'base_subnet.BaseSubnet'):
        """Add network interface to container

        Network interfaces are used to store node's IP address and names of the virtual
        network interfaces that will be created on node start (like br-xxxxxx for docker networks)
        This functions called mostly from Subnet.add_node.
        """
        self._interfaces[ip_address] = subnet

    def crete_gateway(self, from_network: 'Network', to_network: 'Network'):
        """Create a gateway from from_network to to_network on this node

        From_network is always a local network and
        to_network can be global or local.
        Gateway node needs to be connected to at least one subnet
        in each network before calling this function.
        """
        if not from_network.is_local:
            raise Exception('from_network must be local!')

        if from_network not in self._networks or to_network not in self._networks:
            raise Exception('Gateway node needs to be connected to at least one subnet in each network')

        from_network.lan_gateway = self
        self._gateways.append((from_network, to_network))

    def _start(self, *args, **kwargs):
        """Start function to be implemented in resource-specific classed, like DockerNode"""

        raise NotImplementedError()

    def start(self, *args, **kwargs):
        """Start function called from Network.start

        Start procedure starts with gathering routing information from
        node's subnets and LANs. Then Node._start is called to create the node itself,
        after that Node._connect_subnets connects node to it's subnets and
        Node._deploy_routes sets up node's routes.
        """

        if self.started:
            return

        ret = self._start(*args, **kwargs)
        self.started = True
        print('node', self.name, 'started')

        self._setup_routes()
        if self.is_gateway:
            for network in self._networks:
                if self is network.lan_gateway and not network.started:
                    network.start()

        self._connect_subnets()
        self._deploy_routes()
        if self.is_gateway:
            self._setup_lan_gateway()

        return ret

    def _stop(self, *args, **kwargs):
        """Stop function to be implemented in resource-specific classed, like DockerNode"""

        raise NotImplementedError()

    def stop(self, *args, **kwargs):
        """
        Stop function called from Network.start
        """
        if self.stopped:
            return

        self._stop(*args, **kwargs)
        self.stopped = True

        if self.is_gateway:
            for network in self._networks:
                if self is network.lan_gateway:
                    network.stop()

    def _connect_subnets(self):
        for ip, network in self._interfaces.items():
            self._connect_subnet(network, ip)

    def _connect_subnet(self, subnet, ip):
        """Function connecting node resources to virtual subnet to
        be implemented in resource-specific classed, like DockerNode"""
        raise NotImplementedError()

    def _setup_routes(self):
        for network in sorted(self._networks, key=lambda n: n.is_local, reverse=True):

            # for routes we can only find neighbor gateway
            # intentionally not doing this for nodes
            if self.is_router and network.is_local:
                for subnet in self.subnets:
                    if network.lan_gateway in subnet:
                        for ip_range in network.lan_gateway.get_gateway_ranges(network):
                            if ip_range not in self._routes:
                                self._routes[ip_range] = network.lan_gateway.get_ip(subnet)

            # for gateways and nodes find neighbor router (or gateway for plain node)
            elif not self.is_router:

                ip_ranges = ['default', ]
                if network.is_local:
                    if self == network.lan_gateway:
                        ip_ranges = [network.ip_range, ]

                    elif self.connect_to_internet or 'default' in self._routes:
                        ip_ranges = network.lan_gateway.get_gateway_ranges(network)

                elif self.connect_to_internet:
                    ip_ranges = [network.ip_range, ]

                for subnet in self.subnets:
                    if subnet in network:
                        gateway = subnet.get_gateway(self)
                        if gateway:
                            for ip_range in ip_ranges:
                                if ip_range not in self._routes:
                                    self._routes[ip_range] = gateway

    def get_gateway_ranges(self, network):
        """Returns ip ranges for all the networks accessible via this gateway"""
        # TODO: not sure if it works
        if self.is_gateway:
            return sum(
                [
                    to_network.ip_range,
                    *(to_network.lan_gateway.get_gateway_ranges(network) if network.lan_gateway else [])
                ] for from_network, to_network in self._gateways
                if from_network is network and self is not network.lan_gateway
            )
        else:
            return []

    def _deploy_routes(self):
        """Function deploing routing information in node resources
        to be implemented in resource-specific classed, like DockerNode"""
        raise NotImplementedError()

    def _setup_lan_gateway(self):
        """Function setting up LAN gateway on this node
        to be implemented in resource-specific classed, like DockerNode"""
        raise NotImplementedError()

    @property
    def subnets(self) -> List['BaseSubnet']:
        return list(self._interfaces.values())

    @property
    def neighbors(self):
        return sum([subnet.nodes for subnet in self.subnets])

    def add_network(self, net: 'Network'):
        if net in self._networks:
            return
        if not net.is_local and any(filter(lambda n: not n.is_local, self._networks)):
            raise Exception('Node cannot reside in two global networks at the same time!')
        self._networks.append(net)

    def load(self):
        raise NotImplementedError()

    # TODO: proper realisation with Subnet optional argument
    def get_ip(self, net=None):
        if net is None:
            return next(iter(self._interfaces.keys()), None)

        else:
            for ip, network in self._interfaces.items():
                if network == net:
                    return ip

            return None

    @classmethod
    def make_router(cls, *args, **kwargs):
        """Function defining router node to be implemented in resource-specific classed, like DockerNode"""
        raise NotImplementedError()

    @classmethod
    def generate_nodes(cls, name_prefix, amount, *args, **kwargs):
        return [cls(name_prefix + str(i+1), *args, **kwargs) for i in range(amount)]

    @classmethod
    def generate_routers(cls, name_prefix, amount, *args, **kwargs):
        return [cls.make_router(name_prefix + str(i+1), *args, **kwargs) for i in range(amount)]

    def activate_protocol(self, protocol, config):
        """Function activating routing protocol for node
        to be implemented in resource-specific classed, like DockerNode"""
        raise NotImplementedError()

    def __repr__(self):
        return '<{class_name}({name}, {ip})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip=self.get_ip())

    def unlock(self):
        """Unlock node from waiting mode

        Node: WIP

        Starting large networks can take a long time, so if one node workload
        depends on the other node it cant be launched before the other node started.
        If you want all nodes to start simultaneously, a waiting mode before workload
        is required. For DockerNode, for example, we modify entrypoint to make
        original entrypoint and cmd to be launched only if certain file exists and
        than create that file on unlocking.

        """
        # TODO: not the best decision, but we need to avoid starting container before routes are set up
        # temporary bypass

        # print(f'unlocking {self.name}')
        self.put_string(self.lock_filepath, '')

    def put_string(self, path, string):
        """Put sting in file on node"""
        raise NotImplementedError()

    def in_net(self, net):
        return net in self._networks



from .. import base
from . import utils

import docker


class DockerSubnet(base.BaseSubnet):

    _client = utils.docker_client

    def __init__(self, *args, size=1023, docker_client=None, **kwargs):
        self.docker_network = None
        if docker_client:
            self._client = docker_client

        super().__init__(*args, size=size + 1, **kwargs)
        # minus one address for docker gateway
        self._docker_gateway = self._get_host_ip()

    def _make_ipam_config(self):
        ipam_pool = docker.types.IPAMPool(subnet=str(self._ip_range), gateway=self._docker_gateway)
        ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool, ])

        return ipam_config

    def start(self):
        if self.started:
            return self.docker_network

        # for some reason IPAM configuration in docker-py break swarm networks, so we'll use CLI for now
        self.docker_network = self._client.networks.create(
            self.name,
            ipam=self._make_ipam_config())

        self.started = True
        self.loaded = True

        print('[ldc] subnet', self.name, 'started')

        return self.docker_network

    def load(self):
        try:
            self.docker_network = self._client.networks.get(self.name)
            self.loaded = True
        except docker.errors.NotFound:
            pass

    def stop(self):
        if not self.loaded:
            self.load()

        try:
            self.docker_network.remove()
            print('[ldc] subnet', self.name, 'stopped')
        except Exception as e:
            print('error while removing subnet {}: {}'.format(self.name, e))



import l3ns.base.network


network = l3ns.base.network.Network('22.0.0.0/16')
"""Default network for all Subnets"""

import l3ns.ldc.subnet
subnet_class = l3ns.ldc.subnet.DockerSubnet
"""Default subnet class for Node.connect_to and Network.create_subnet"""




import docker

dc = docker.from_env()


def show_taken_networks():

    for net in dc.networks.list():
        try:
            print('{}: {}'.format(net.name, net.attrs['IPAM']['Config'][0]['Subnet']))
        except (KeyError, IndexError):
            print('{}: Error: {}'.format(net.name, 'No config'))
        except Exception as e:
            print('{}: Error: {}'.format(net.name, e))


if __name__ == "__main__":
    show_taken_networks()



from ipaddress import ip_network, IPv4Network
from math import log2, ceil
from typing import Union, Optional, TYPE_CHECKING, List
from .. import defaults

import concurrent.futures
import traceback

import time

if TYPE_CHECKING:
    from .node import BaseNode
    from .subnet import BaseSubnet


class Network:
    """Base class for forming local and wide area networks"""

    def __init__(self, ip_range: Union[str, IPv4Network], local=False):
        """Creates network with given ip range.

        Creates local or wide area nerwork with a given ip range that will be
        distributed to subnetworks that will be created in this network.

        Args:
            ip_range: IP address range for the networks
            local: Optional; Set to True to create local network

        """
        self.ip_range = str(ip_range)
        self._available_subnets = [ip_network(ip_range) if type(ip_range) is str else ip_range, ]
        self._subnets: List['BaseSubnet'] = []
        self._nodes: List['BaseNode'] = []
        self.is_local = local
        self.lan_gateway: Optional['BaseNode'] = None

        self.loaded = False
        self.started = False

    def start(self, interactive=False):
        """Launches all the subnetworks and nodes in this network

        This function calls appropriate functions for subnets first
        and for nodes later. Even nodes that are not connected to
        any subnets will be started.

        Interactive mode allows user to start and stop the network
        with a simpler script, locking at the shell prompt
        after the network is started and shutting it down on user
        input. Non-interactive mode doesn't lock code execution, so
        you can create more complex testing scripts.


        Args:
             interactive: Optional; Sets whether interactive mode will be used

        """
        if self.loaded:
            raise Exception('Network already loaded!')

        if not interactive:
            for subnet in self._subnets:
                subnet.start()

            for node in self._nodes:
                if node is not self.lan_gateway:
                    node.start()

            for node in self._nodes:
                if node is not self.lan_gateway:
                    node.unlock()

            self.started = True
            self.loaded = True

        else:
            try:
                self.start(interactive=False)
            except:
                print('\nException while starting:\n_________________\n\n',
                      traceback.format_exc(), '\n_________________\n', sep='')

            i = ''
            while i not in ('y', 'n', 'yes', 'no'):
                i = input('Stop? (y/n): ')

            if i[0] == 'y':
                self.stop()

    def load(self, *args):
        """Loads network resources

        Loads network resources (i.e. containers and
        docker networks in l3ns.ldc) that were previously
        launched in a different script back in the memory.


        Note that this will not load network stricture, it
        needs to be already defined. This allows you to work
        with network resources from several scripts.

        """
        for node in self._nodes:
            node.load(*args)

        for subnet in self._subnets:
            subnet.load(*args)

    def stop(self, *args):
        """Stops the network"""
        if not self.loaded:
            self.load(*args)

        for node in self._nodes:
            if node is not self.lan_gateway:
                node.stop(*args)

        for subnet in self._subnets:
            subnet.stop(*args)

        self.started = False

    def __contains__(self, item):
        return item in self._nodes or item in self._subnets

    def add_node(self, node: 'BaseNode'):
        """Add node to the network"""
        if not node.in_net(node):
            self._nodes.append(node)
        node.add_network(self)

    def add_subnet(self, subnet):
        """Add subnet to the network"""
        if subnet not in self._subnets:
            self._subnets.append(subnet)

    def create_subnet(self, subnet_name, *args, subnet_type: type = None, **kwargs):
        """Create subnet in this network"""
        if subnet_type is None:
            subnet_type = defaults.subnet_class

        return subnet_type(subnet_name, *args, network=self, **kwargs)

    def get_subnet_range(self, size: int):
        """Retrieve an IP range of given size for subnet and remove it from available addresses"""
        max_prefix = 32 - ceil(log2(size + 2))
        #  size = 2^(32-p) - 2

        try:
            subnet = [net for net in self._available_subnets if net.prefixlen <= max_prefix][0]
        except IndexError:
            raise Exception('Network address pool is empty')

        self._available_subnets.remove(subnet)

        if subnet.prefixlen == max_prefix:
            return subnet

        try:
            smaller_subnet = next(subnet.subnets(new_prefix=max_prefix))
        except ValueError or StopIteration as e:
            raise Exception('Error while splitting net: {}'.format(e))

        self._available_subnets.extend(subnet.address_exclude(smaller_subnet))
        self._available_subnets.sort(key=lambda n: 32 - n.prefixlen)

        return smaller_subnet

    def connect_lan(self, lan_network: 'Network', gateway: 'BaseNode'):
        return gateway.crete_gateway(lan_network, self)

    def __enter__(self):
        self.start()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()


class NetworkConcurrent(Network):
    """WIP-interface to launch large networks (500+ nodes) concurrently."""

    def __init__(self, *args, max_workers=None, **kwargs):
        self.max_workers = max_workers
        super().__init__(*args, **kwargs)

    def start(self, interactive=False):
        """Starts the network concurrently using ThreadPoolExecutor"""
        # TODO: debug & interactive realisation

        if self.loaded:
            raise Exception('Network already loaded!')

        if not interactive:
            for subnet in self._subnets:
                subnet.start()

            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                for node in self._nodes:
                    executor.submit(node.start, dc=None)

            print('start time:', time.strftime('[%H:%M:%S]', time.gmtime()))

            # TODO: executor here too
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                for node in self._nodes:
                    executor.submit(node.unlock)

            for node in self._nodes:
                node.unlock()

            self.started = True
            self.loaded = True

        else:
            try:
                self.start(interactive=False)
            except:
                print('\nException while starting:\n_________________\n\n',
                      traceback.format_exc(), '\n_________________\n', sep='')

            i = ''
            while i not in ('y', 'n', 'yes', 'no'):
                i = input('Stop? (y/n): ')

            if i[0] == 'y':
                self.stop()

    def stop(self):
        """Stops the network concurrently using ThreadPoolExecutor"""
        if not self.loaded:
            self.load()

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            for node in self._nodes:
                executor.submit(node.stop, dc=None)

        for subnet in self._subnets:
            subnet.stop()

        self.started = False



from .. import defaults
from .. import utils
from . import network as base_network
from . import node as base_node

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .network import Network

class BaseSubnet:
    """Base class for subnets representing IP subnets of the virtual networks"""

    def __init__(self, name: str, *args, size: int = 1024, network: 'base_network.Network' = None):
        """Create subnet with given name and size in given Network

        Names of networks need to be unique, as with Nodes.
        Size can't be anything as long as Network has free IP addresses in pool.
        However there is to caveats:
        1. Some implementation must have several IP addresses reserved, for example
        DockerNetwork reserves one IP address for virtual bridge interface.
        This will increase network size.
        2. Actual subnet IP address size must be a power of two. So actual size of network will be
        increase to a nearest power of two.

        Args:
            name: name for new subnet
            *args: nodes to add to the new network
            size: Optional; size of a new network
            network: Optional; as the name implies every subnet lies in some network.
                     If network is not provided, l3ns.defaults.network will be used.
        """
        self.name = name
        self._network = network if network is not None else defaults.network
        self._network.add_subnet(self)
        self._max_size = size
        self._ip_range = self._network.get_subnet_range(self._max_size)
        self._hosts = list(self._ip_range.hosts())

        self._nodes_dict = {}
        self.started = False
        self.loaded = False

        for node in utils.args.list_from_args(args):
            self.add_node(node)

    def __iter__(self):
        return iter(self._nodes_dict)

    def _get_host_ip(self):
        try:
            return str(self._hosts.pop(0))
        except IndexError:
            raise Exception('Network {} is too small'.format(self.name))

    def add_node(self, node: 'base_node.BaseNode'):
        """Add node to the network

        This is a main method to define network structure.
        This will also add node to subnet's network, create interface in node with an
        IP address from this subnet.
        """

        ip_address = self._get_host_ip()

        self._nodes_dict[ip_address] = node
        if node not in self._network:
            self._network.add_node(node)
        node.add_interface(ip_address, self)

    def start(self):
        """Subnet starting procedure to be implemented in resource-specific classed, like DockerSubnet"""
        raise NotImplementedError()

    def load(self):
        """Subnet loading procedure to be implemented in resource-specific classed, like DockerSubnet"""
        raise NotImplementedError()

    def stop(self):
        """Subnet stopping procedure to be implemented in resource-specific classed, like DockerSubnet"""
        raise NotImplementedError()

    def get_network(self) -> 'Network':
        """Get Network this subnet resides in."""
        return self._network

    def get_gateway(self, node):
        """Get default gateway for this network

        Returns default gateway (not LAN gateway) for nodes in this subnet
        """
        router = None

        for ip, n in self._nodes_dict.items():

            if n is node:
                continue

            if not router and n.is_router:
                router = ip

        return router

    def get_nodes_dict(self):
        """Get nodes in a dict with IP addresses as keys"""
        return self._nodes_dict.copy()

    def prefixlen(self):
        """Get subnet IP address prefix length"""
        return self._ip_range.prefixlen

    def get_ip_range(self):
        """Get subnet IP address,"""
        return self._ip_range

    @property
    def nodes(self):
        """Get a list of subnet's nodes"""
        return list(self._nodes_dict.values())

    def get_node_ip(self, node):
        for ip, n in self._nodes_dict.items():
            if n is node:
                return ip

        raise Exception('Node is not in subnet')

    def __repr__(self):
        return '{class_name}({name}, {ip_range})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip_range=self._ip_range
        )

    def __contains__(self, item):
        return item in self._nodes_dict.values()

    def __str__(self):
        return str(self._ip_range)



import os
import subprocess
from collections import defaultdict
from typing import List, TYPE_CHECKING
from tempfile import TemporaryDirectory

from ..base.network import Network


if TYPE_CHECKING:
    from ..swarm.node import SwarmNode
    from ..swarm.utils import SwarmHost


class AutoClusterNetwork(Network):
    """Can be used only with AutoSwarmNode and AutoSwarmSubnet

    Local networks are not supported yet.
    Fixed nodes also are not supported yet, sorry.
    """
    _nodes: List['SwarmNode']

    # TODO: store preset in this module and find out abs path dynamically
    def __init__(
            self, ip_range, hosts: List['SwarmHost'], *args,
            kahypar_preset='/home/panda/kahypar/config/cut_kKaHyPar_sea20.ini',
            max_imbalance=0.03,
            **kwargs
    ):
        self.hosts = hosts
        self.kahypar_preset = kahypar_preset
        self.max_imbalance = max_imbalance
        super().__init__(ip_range, *args, **kwargs)

    def split_network(self):
        """This function will assign a cluster host for each Node in network"""
        subnet_to_node_idx = defaultdict(list)
        fixed_nodes_file_lines = []

        for node_idx, node in enumerate(self._nodes):
            for subnet in node.subnets:
                subnet_to_node_idx[subnet].append(str(node_idx+1))
                # fixed_nodes_file_lines.append('0' if node.is_router else '-1')
                fixed_nodes_file_lines.append('-1')

        hMetis_file_lines = [
            '{} {}'.format(len(self._subnets), len(self._nodes)),
            *[' '.join(subnet_to_node_idx[subnet]) for subnet in self._subnets]
        ]

        with TemporaryDirectory() as tmp_dir:
            hMetis_file = os.path.join(tmp_dir, 'in.graph')
            fixed_nodes_file = os.path.join(tmp_dir, 'fix_file')

            with open(hMetis_file, 'w') as tmp_file:
                tmp_file.write('\n'.join(hMetis_file_lines) + '\n')

            with open(fixed_nodes_file, 'w') as tmp_file:
                tmp_file.write('\n'.join(fixed_nodes_file_lines) + '\n')

            ret = subprocess.run(
                'kahypar -h {0} -k {1} -e {2} -o cut -m direct -f {4} -p {3} -w True'.format(
                    hMetis_file,
                    len(self.hosts),
                    self.max_imbalance,
                    self.kahypar_preset,
                    fixed_nodes_file
                ),
                shell=True,
                capture_output=True)

            if ret.returncode != 0:
                raise Exception("Error while using KaHyPar: \n" + ret.stdout.decode() + '\n' + ret.stderr.decode())

            try:
                ret_filename = os.path.join(
                    tmp_dir,
                    next(filter(
                            lambda f: f not in [
                                os.path.basename(hMetis_file),
                                os.path.basename(fixed_nodes_file)
                            ],
                            os.listdir(tmp_dir)
                    )))
            except StopIteration:
                raise Exception("Error: KaHYPar did not create any files")

            with open(ret_filename) as ret_file:
                raw_partition = [int(line) for line in ret_file.readlines()]

        for node_idx, part_idx in enumerate(raw_partition):
            self._nodes[node_idx].cluster_host = self.hosts[part_idx]

    def start(self, interactive=False):
        self.split_network()
        return super().start(interactive=interactive)

    def load(self):
        self.split_network()
        return super().load()



from ..swarm.subnet import SwarmSubnet
from ..swarm.node import SwarmNode
from .network import AutoClusterNetwork

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..swarm.utils import SwarmHost


class AutoSwarmSubnet(SwarmSubnet):
    _network: AutoClusterNetwork

    def add_node(self, node: 'SwarmNode'):

        ip_address = self._get_host_ip()

        # docker swarm overlay requires additional ip address for lode balancer each cluster node (cluster_host in l3ns)
        if len(self._cluster_hosts) < len(self._network.hosts):
            self._cluster_hosts[self._network.hosts[len(self._cluster_hosts)]] = self._hosts.pop(0)

        self._nodes_dict[ip_address] = node
        if node not in self._network:
            self._network.add_node(node)
        node.add_interface(ip_address, self)

    def start(self):
        if self.started:
            return self.docker_network

        one_host = self.is_one_host()
        if one_host:
            self._client = one_host.get_docker_client()

        self.docker_network = self._client.networks.create(
            self.name,
            driver='overlay' if not one_host else None,
            attachable=True,
            ipam=self._make_ipam_config())

        if not self.docker_network.attrs['Driver']:
            raise Exception('Error: IPAM customisation failed for swarm network {}, check if other networks overlap ip pool'.format(self.name))

        self.started = True
        self.loaded = True

        print('[swarm] subnet', self.name, 'started')

        return self.docker_network

    def load(self):
        one_host = self.is_one_host()
        if one_host:
            self._client = one_host.get_docker_client()

        return super().load()

    def is_one_host(self) -> Union[bool, 'SwarmHost']:
        nodes = self.nodes
        first = nodes[0].cluster_host
        return first if all(n.cluster_host is first for n in nodes) else False



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='l3ns-dev',
    version='0.0.2',
    license='MIT',
    description='Simple Docker based Network Simulator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Daniil Malevanniy and others',
    author_email='oleg.Jakushkin@gmail.com',
    url='https://github.com/OlegJakushkin/l3ns',
    packages=setuptools.find_packages(),
    install_requires=[
          'docker',
          'hashids',
          'pydoc-markdown'
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        
    ],
)



from .. import ldc
from .utils import SwarmHost


class SwarmNode(ldc.DockerNode):

    def __init__(self, name, cluster_host: 'SwarmHost', **kwargs):
        self.cluster_host = cluster_host

        super().__init__(name, **kwargs)
        self._client = None

    # TODO: move client acquisition to runtime in DockerNode too?
    def _start(self):
        self._client = self.cluster_host.get_docker_client()
        return super()._start()

    def load(self):
        self._client = self.cluster_host.get_docker_client()
        return super().load()

    def _connect_subnet(self, subnet, ip):
        self._client.api.connect_container_to_network(self.container.name, subnet.docker_network.name, ipv4_address=ip)

    def __repr__(self):
        return '<{class_name}({name}@{hostname}, {ip})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip=self.get_ip(),
            hostname=self.cluster_host.address)
